//---------------------------------------------------------------------------
// File: TMainForm.cpp, TDx_Draw_Library, Example8 - Page Flipping
//
// Date: 21st June, 2002
// Author: Darren John Dwyer
// Copyright: (c) 2002 Darren John Dwyer, All Rights Reserved.
//
// Description:
//
// This example shows how to create a primary surface and backbuffer, then
// uses a standard TTimer::OnTick() to perform simple page-flipping.
//
// Please note that the page-flipping is relatively slow for 3 reasons :-
//
// (1) the TTimer is very slow
// (2) the entire backbuffer is being repainted each frame using slow vcl paints.
// (3) a whole bunch of stuff has not been optimized
//
// See later examples for increased frame rates.
//
//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop
//---------------------------------------------------------------------------
#include "TMainForm.h"
//---------------------------------------------------------------------------
#if (__BORLANDC__ >= 0x0530) // BCB Version 3 +
    #pragma package(smart_init)
#endif
//---------------------------------------------------------------------------
#pragma link "tddscaps"
#pragma link "tddsurfacedesc"
#pragma link "tdx_draw"
#pragma link "tdx_drawsurface"
#pragma resource "*.dfm"
//---------------------------------------------------------------------------
#pragma link "TDx_Draw"
#pragma link "TDx_DrawSurface"
#pragma link "TDDCaps"
#pragma link "TDDPixelFormat"
#pragma link "TDDSCaps"
#pragma link "TDDSurfaceDesc"
#pragma link "TDx_Draw_Library_Install"
//---------------------------------------------------------------------------
TMainForm *MainForm;
//---------------------------------------------------------------------------
AnsiString IntToHexStr( unsigned int hexvalue )
{
char* hexchars[16] = {"0","1","2","3","4","5","6","7","8","9","A","B","C","D","E","F"};
AnsiString string = "0x";
int value = 0x10000000;
for (int i=0;i<8;i++)
    {
    string += hexchars[hexvalue/value];
    hexvalue -= value * (hexvalue/value);
    value = value >> 4;
    }
return string;
}
//---------------------------------------------------------------------------
#define Msg(A)  {\
                if (StatusMemo->Text.Length()>30000) StatusMemo->Text = "";\
                StatusMemo->Lines->Add(A);\
                }
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
    : TForm(Owner)
{
canvas = new TCanvas();
}
//---------------------------------------------------------------------------
__fastcall TMainForm::~TMainForm()
{
if (canvas) delete canvas;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormCreate(TObject *Sender)
{
Top=0;
Left=0;
Width=800;
Height=600;
CurrentPageNumber = 0;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormDestroy(TObject *Sender)
{
if (Dx_Draw1->Created) Dx_Draw1->Destroy();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ExitButtonClick(TObject *Sender)
{
Close();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ClearButtonClick(TObject *Sender)
{
StatusMemo->Lines->Clear();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_Draw1Create(TObject *Sender)
{
NumModes=0;
ModesListBox->Items->Clear();
Dx_Draw1->EnumDisplayModes(0,NULL,NULL);

Dx_Draw1->SetCooperativeLevel( Application->Handle, DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE );

// create a complex primary surface
PrimarySurfaceDesc->Flags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
PrimarySurfaceDesc->BackBufferCount = 1;
PrimaryCaps->Caps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP;
PrimarySurfaceDesc->SCaps = PrimaryCaps;
PrimarySurface->Create( PrimarySurfaceDesc, Dx_Draw1 );
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_Draw1Destroy(TObject *Sender)
{
PrimarySurface->Destroy();
Dx_Draw1->SetCooperativeLevel( Application->Handle, DDSCL_NORMAL );
if (Dx_Draw1->RestoreDisplayMode())
    {
    Top=0;
    Left=0;
    Width=800;
    Height=600;
    Invalidate();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_Draw1Error(TObject *Sender, AnsiString Function,
    AnsiString Error, AnsiString ErrorMessage)
{
Msg("ERROR ["+Error+"]: "+Function+", "+ErrorMessage);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::Dx_Draw1EnumDisplayModes(TObject *Sender,
      TDDSurfaceDesc *SurfaceDesc, void *Context, bool &Finished)
{
int width = SurfaceDesc->Width;
int height = SurfaceDesc->Height;
int refresh_rate = SurfaceDesc->RefreshRate;
int bit_depth;

if (SurfaceDesc->PixelFormat->Flags[0] & DDPF_PALETTEINDEXED1) bit_depth = 1;
else if (SurfaceDesc->PixelFormat->Flags[0] & DDPF_PALETTEINDEXED2) bit_depth = 2;
else if (SurfaceDesc->PixelFormat->Flags[0] & DDPF_PALETTEINDEXED4) bit_depth = 4;
else if (SurfaceDesc->PixelFormat->Flags[0] & DDPF_PALETTEINDEXED8) bit_depth = 8;
else if (SurfaceDesc->PixelFormat->Flags[0] & DDPF_PALETTEINDEXEDTO8) bit_depth = 8;
else if (SurfaceDesc->PixelFormat->Flags[0] & DDPF_ALPHA) bit_depth = SurfaceDesc->PixelFormat->AlphaBitDepth[0];
else if (SurfaceDesc->PixelFormat->Flags[0] & DDPF_ALPHAPIXELS) bit_depth = SurfaceDesc->PixelFormat->AlphaBitDepth[0];
else if (SurfaceDesc->PixelFormat->Flags[0] & DDPF_ALPHAPREMULT) bit_depth = SurfaceDesc->PixelFormat->AlphaBitDepth[0];
else if (SurfaceDesc->PixelFormat->Flags[0] & DDPF_BUMPDUDV) bit_depth = SurfaceDesc->PixelFormat->BumpBitCount[0];
else if (SurfaceDesc->PixelFormat->Flags[0] & DDPF_BUMPLUMINANCE) bit_depth = SurfaceDesc->PixelFormat->BumpBitCount[0];
else if (SurfaceDesc->PixelFormat->Flags[0] & DDPF_RGB) bit_depth = SurfaceDesc->PixelFormat->RGBBitCount[0];
else if (SurfaceDesc->PixelFormat->Flags[0] & DDPF_YUV) bit_depth = SurfaceDesc->PixelFormat->YUVBitCount[0];

if (NumModes<MAX_MODES)
    {
    Widths[NumModes] = width;
    Heights[NumModes] = height;
    BitDepths[NumModes] = bit_depth;
    RefreshRates[NumModes] = refresh_rate;
    NumModes++;

    ModesListBox->Items->Add( IntToStr(width)+"x"+IntToStr(height)+"x"+IntToStr(bit_depth)+" @ "+IntToStr(refresh_rate)+"Hz" );
    }
else
    Finished = true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::NextModeButtonClick(TObject *Sender)
{
if (ModesListBox->Items->Count>0)
    {
    Msg("Selecting Next Available Mode");
    if (ModesListBox->ItemIndex==ModesListBox->Items->Count-1) ModesListBox->ItemIndex=0;
    else ModesListBox->ItemIndex = ModesListBox->ItemIndex + 1;
    ModesListBox->OnClick(this);
    }
else
    ModesListBox->ItemIndex = -1;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ModesListBoxClick(TObject *Sender)
{
if (ModesListBox->ItemIndex!=-1)
    {
    int index = ModesListBox->ItemIndex;
    AnsiString name = ModesListBox->Items->Strings[index];

    Msg("Mode Selected: '"+name+"'");
    CurrentModeLabel->Caption = name;
    }
else
    CurrentModeLabel->Caption = "-- default --";
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::PrimarySurfaceCreate(TObject *Sender)
{
Msg(">>> Primary Surface Created");

// get it's attached backbuffer
if (PrimarySurface->GetAttachedSurface( BackBufferCaps, BackBuffer ))
    // call the backbuffer::OnCreate() event manually.
    BackBuffer->OnCreate(this);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::PrimarySurfaceDestroy(TObject *Sender)
{
Msg(">>> Primary Surface Destroy");
if (BackBuffer->Created) BackBuffer->Destroy();
}
//---------------------------------------------------------------------------




void __fastcall TMainForm::PrimarySurfaceSurfaceLost(TObject *Sender)
{
PrimarySurface->Restore();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ModesListBoxDblClick(TObject *Sender)
{
SelectModeButton->OnClick(this);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SelectModeButtonClick(TObject *Sender)
{
if (ModesListBox->Items->Count>0)
    {
    ShowMessage("Please Note:\n\
You may experience difficulty restoring your display mode.\n\
To return to the original display mode at any time, press ALT-R (Restore) or ALT-X (eXit)" );

    int index = ModesListBox->ItemIndex;
    PrimarySurface->Destroy();
    Dx_Draw1->SetDisplayMode( Widths[index], Heights[index], BitDepths[index], RefreshRates[index], 0 );
    PrimarySurface->Create( PrimarySurfaceDesc, Dx_Draw1 );

    RepaintButton->OnClick(this);
    }
}
//---------------------------------------------------------------------------


void __fastcall TMainForm::RepaintButtonClick(TObject *Sender)
{
Timer1->Enabled = false;
if (Dx_Draw1->Created)
	{
    Dx_Draw1->FlipToGDISurface();
    CurrentPageNumber = 0;
    }

int old_top = Top;
Top=Screen->Height+1;
Invalidate();
Top=old_top;
Invalidate();
}
//---------------------------------------------------------------------------



void __fastcall TMainForm::BackBufferCreate(TObject *Sender)
{
Msg(">>> BackBuffer Created");
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::BackBufferDestroy(TObject *Sender)
{
Msg(">>> BackBuffer Destroyed");
}
//---------------------------------------------------------------------------


void __fastcall TMainForm::CreateButtonClick(TObject *Sender)
{
Dx_Draw1->Create(NULL);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FlipButtonClick(TObject *Sender)
{
if (Timer1->Enabled) return;
if (!BackBuffer->Created) return;
if (!PrimarySurface->Created) return;

ShowMessage("Please Note:\n\
After the primary surface has been painted the contents of the current form\n\
will no longer be visible though it is still active. To repaint the form\n\
Press ALT-R (Restore) or to exit the application, press ALT-X (eXit)" );

Timer1->Enabled = true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Timer1Timer(TObject *Sender)
{
if (!BackBuffer->Created) return;
if (!PrimarySurface->Created) return;
// create a canvas to use for drawing directly onto the back buffer
if (canvas==NULL)
    {
    Timer1->Enabled = false;
    ShowMessage( "Canvas is NULL" );
    return;
    }

// error-proof vcl code
HDC hdc;
if (BackBuffer->GetDC(&hdc))
    {
    // paint directly onto the backbuffer using VCL paint code via a temporary TCanvas
    try
        {
        canvas->Handle = hdc;

        // paint the background star image
        canvas->StretchDraw( canvas->ClipRect, Image1->Picture->Graphic );

    	// blit the contents of the form to the backbuffer, a little bit tricky
		//
        // In order to paint the current form, you need to tell bcb to paint the form
        // to a particular device context, which is fair enough, you can generally
        // get away with using the TForm::PaintTo() function.
        //
        // Problem is, the PaintTo() only paints the client area of the form, and
        // it doesn't pickup on things such as carets.
        //
        // So, you need to also paint the frame part of the form, which can be done
        // using the win32 function RedrawWindow()
        //
        // Problem is, there appears to be no way to tell windows to paint the
        // frame of a window anywhere other than to the standard windows gdi device context.
        //
        // --- Anyone know of a way to tell the GDI to paint a window frame to a user-supplied device context? ---
        // -- if so, please email details to :- admin@bcb-tools.com
        //
        // So, you can paint the window frame to the current gdi device context, which is
        // either the backbuffer or the primary surface, depending on how many times
        // PrimarySurface->Flip() has been called.
        //
        // If the windows gdi device context is currently the backbuffer, the work is already done.
        // If the windows gdi device context is currently the primary surface, you need to copy it to the backbuffer.
        //
        // When copying, attempting to HDC hdc = GetDC( NULL ); does not work.
        // Similarly, attempting to Dx_Draw1->GetGDISurface( Temp_DrawSurface1 ); HDC hdc; Temp_DrawSurface1->GetDC( hdc ); does not work.
        //
        // The following does appear to work - mostly (except for the frames sometimes flicker).
		//
        // --
        //
        // Implementation is as follows :-
        //
        // * paint the background image (implemented above this comment block)
        // * paint the form's window frame to the windows gdi surface
        // * if the gdi output is currently visible on the monitor... copy it to the backbuffer
        // * copy the contents of the form's client area to the backbuffer
        // * flip the backbuffer to the primary surface
        //
        // --
        //
        // Alas, if only the TDx_Draw_Form (the revamped old TDirectDrawForm) was fully working...
        //
        // --

       	// paint the frame to the gdi surface (either the primary surface or backbuffer)
		RedrawWindow( Handle, NULL, NULL, RDW_FRAME | RDW_INVALIDATE );

       	// if the gdi surface is currently visible on the monitor (the gdi and primary surface are the same)
        if (CurrentPageNumber%2==0)
        	{
        	// copy the recently painted window frame from the current gdi device context to the backbuffer
            HDC frame_dc = GetWindowDC( Handle );
	        if (frame_dc)
	        	{
                int width = max( Width, Screen->Width );
                int height = max( Height, Screen->Height );
	        	BitBlt( hdc, Left, Top, width, height, frame_dc, Left, Top, SRCCOPY );
				ReleaseDC( Handle, frame_dc );
	            }
            }

        // paint everything inside the form client area directly to the backbuffer
        TPoint client = { 0, 0 };
        TPoint screen = ClientToScreen( client );
        PaintTo( hdc, screen.x, screen.y );

        // paint some text
        canvas->Brush->Style = bsClear;
        canvas->Font->Name = "Arial";
        canvas->Font->Style = TFontStyles();
        canvas->Font->Size = 20;
        canvas->Font->Color = (TColor) RGB(0,0,0);
        canvas->TextOut( screen.x+2, screen.y+2, "Paint onto the BackBuffer, then flip to the Primary Surface" );
        canvas->Font->Color = (TColor) RGB(128,255,128);
        canvas->TextOut( screen.x, screen.y, "Paint onto the BackBuffer, then flip to the Primary Surface" );

        // and more text
        canvas->Brush->Style = bsClear;
        canvas->Font->Name = "Comic Sans MS";
        canvas->Font->Style = TFontStyles() << fsBold << fsItalic << fsUnderline << fsStrikeOut;
        canvas->Font->Size = 24;
        canvas->Font->Color = (TColor) RGB(0,0,0);
        canvas->TextOut( screen.x+2+CurrentPageNumber%Screen->Width/2, screen.y+32, "Frame# "+IntToStr(CurrentPageNumber) );
        canvas->Font->Color = (TColor) RGB(128+CurrentPageNumber%128,255,128+CurrentPageNumber%128);
        canvas->TextOut( screen.x+CurrentPageNumber%Screen->Width/2, screen.y+30, "Frame# "+IntToStr(CurrentPageNumber) );
        }
    catch (Exception &e)
        {
        Timer1->Enabled = false;
        }

    canvas->Handle = NULL;

    // finished with backbuffer
    BackBuffer->ReleaseDC(hdc);
    }

PrimarySurface->Flip( BackBuffer, DDFLIP_DONOTWAIT );
CurrentPageNumber++;
}
//---------------------------------------------------------------------------

